#!/usr/bin/env python3

# Copyright (C) 2021 Gabriele Bozzola
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation; either version 3 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, see <https://www.gnu.org/licenses/>.

"""The :py:mod:`~.twopunctures` module provides infrastructure to prepare
binary black simulations.

Some functions are provided:
- :py:func:`~.prepare_quasicircular_inspiral`, which returns a :py:class:`~.TwoPunctures`
  with parameters set up to lead to a quasi-circular inspiral. The linear momenta
  required for this are computed with a 3.5PN expansion by NRPyPN.

"""

from functools import lru_cache

from jhuki.externals.nrpypn import compute_quasicircular_momenta

# TODO: This module needs unit testing


def prepare_quasicircular_inspiral(
    mass_ratio, coordinate_distance, chi_plus=(0,0,0), chi_minus=(0,0,0)
):
    """

    :param mass_ratio: Mass ratio of the binary.
    :type mass_ratio: float
    :param coordinate_distance: Initial coordinate separation.
    :type coordinate_distance: float

    :param chi_plus: Dimensionless spin of the black hole on the positive side of the x
                     (or z) axis along the three directions.
    :type chi_plus: tuple/list with three numbers

    :param chi_minus: Dimensionless spin of the black hole on the negative side of the x
                     (or z) axis along the three directions.
    :type chi_minus: tuple/list with three numbers

    :returns: A :py:class:`~.TwoPunctures` for a quasi-circular inspiral.
    :rtype: :py:class:`~.TwoPunctures`
    """

    if mass_ratio < 1:
        mass_ratio = 1 / mass_ratio

    Pr, Pt = compute_quasicircular_momenta(
        mass_ratio, coordinate_distance, chi_plus, chi_minus
    )

    momenta_plus = (-Pr, 0, Pt)
    momenta_minus = (Pr, 0, 0 - Pt)

    mass_plus = mass_ratio / (1 + mass_ratio)
    mass_minus = 1 / (1 + mass_ratio)

    return TwoPunctures(
        mass_plus,
        mass_minus,
        coordinate_distance,
        momenta_plus=momenta_plus,
        momenta_minus=momenta_minus,
        chi_plus=chi_plus,
        chi_minus=chi_minus,
    )


class TwoPunctures:
    """The :py:class:`~.TwoPunctures` class contains all the information related to
    one or two black holes, their locations, and properties.

    This class can set up parameters for the ``TwoPunctures`` thorn making sure that
    the center of mass is in the center of the grid (unless differently specified).

    This class is immutable.

    :ivar mass_plus: Mass of the black hole on the positive side of the x (or z) axis.
    :vartype mass_plus: float
    :ivar mass_minus: Mass of the black hole on the negative side of the x (or z) axis.
    :vartype mass_minus: float
    :ivar coordinate_separation: Initial coordinate separation.
    :vartype coordinate_separation: float

    :ivar coord_x_plus: Initial coordinate location of the center of the black hole on the
                        positive side of the x (or z) axis.
    :vartype coord_x_plus: float
    :ivar coord_x_minus: Initial coordinate location of the center of the black hole on the
                         negative side of the x (or z) axis.
    :vartype coord_x_minus: float

    :ivar momenta_plus: Array with the linear momenta along the three directions for the
                        black hole on the positive side of the x (or z) axis.
    :vartype momenta_plus: list/tuple
    :ivar momenta_minus: Array with the linear momenta along the three directions for the
                         black hole on the positive side of the x (or z) axis.
    :vartype momenta_minus: list/tuple

    :ivar chi_plus: Array with the dimensionless_spin along the three directions for the
                    black hole on the positive side of the x (or z) axis.
    :vartype chi_plus: list/tuple
    :ivar chi_minus: Array with the dimensionless spin along the three directions for the
                     black hole on the positive side of the x (or z) axis.
    :vartype chi_minus: list/tuple

    :ivar S_plus: Array with the angular momentum along the three directions for the
                  black hole on the positive side of the x (or z) axis.
    :vartype S_plus: list/tuple
    :ivar S_minus: Array with the angular momentum along the three directions for the
                   black hole on the positive side of the x (or z) axis.
    :vartype S_minus: list/tuple

    :ivar center_offset: Move the geometric center of the system by this amount.
    :vartype center_offset: list/tuple
    """

    def __init__(
        self,
        mass_plus,
        mass_minus,
        coordinate_distance,
        momenta_plus=None,
        momenta_minus=None,
        chi_plus=None,
        chi_minus=None,
        center_offset=None,
    ):
        """Constructor.

        :param mass_plus: Mass of the black hole on the positive side of the x (or z) axis.
        :type mass_plus: float
        :param mass_minus: Mass of the black hole on the negative side of the x (or z) axis.
        :type mass_minus: float
        :param coordinate_separation: Initial coordinate separation.
        :type coordinate_separation: float
        :param par_b: Half of the initial separation.
        :type par_b: float

        :param momenta_plus: Array with the linear momenta along the three directions for the
                            black hole on the positive side of the x (or z) axis.
        :type momenta_plus: list/tuple
        :param momenta_minus: Array with the linear momenta along the three directions for the
                             black hole on the positive side of the x (or z) axis.
        :type momenta_minus: list/tuple

        :param chi_plus: Array with the dimensionless_spin along the three directions for the
                        black hole on the positive side of the x (or z) axis.
        :type chi_plus: list/tuple
        :param chi_minus: Array with the dimensionless spin along the three directions for the
                         black hole on the positive side of the x (or z) axis.
        :type chi_minus: list/tuple

        :ivar center_offset: Move the center of mass of the system by this amount. If None, it
                             this is set so that the center of mass is at [0,0,0].
        :vartype center_offset: list/tuple
        """

        # TODO: Support swapping x and z axis

        def _par_or_zeros(par):
            # Return the parameter if it is not None, otherwise zeros
            return par if par is not None else (0.0,0.0,0.0)

        self.mass_plus = mass_plus
        self.mass_minus = mass_minus
        self.coordinate_distance = coordinate_distance

        self.par_b = coordinate_distance / 2

        self.coord_x_plus = self.coordinate_distance * self.mass_minus
        self.coord_x_minus = self.coordinate_distance * self.mass_plus

        self.momenta_plus = _par_or_zeros(momenta_plus)
        self.momenta_minus = _par_or_zeros(momenta_minus)
        self.chi_plus = _par_or_zeros(chi_plus)
        self.chi_minus = _par_or_zeros(chi_minus)

        self.S_plus = tuple(cp * self.mass_plus ** 2 for cp in self.chi_plus)
        self.S_minus = tuple(cm * self.mass_minus ** 2 for cm in self.chi_plus)

        if center_offset is None:
            self.center_offset = (self.coord_x_plus - self.par_b, 0.0, 0.0)
        else:
            self.center_offset = center_offset

    @property
    @lru_cache(1)
    def parfile_code(self):
        """Return the code you would put in your parfile."""

        def assign_parameter(param, value, which_bh=None, vector=False):
            if which_bh is None:
                return f"TwoPunctures::{param} = {value}"

            # which_bh is either _plus or _minus
            if vector is None:
                return f"TwoPunctures::{param}{which_bh} = {value}"

            if vector:
                return "\n".join([
                    f"TwoPunctures::{param}{which_bh}[{index}] = {value[index]}"
                    for index in range(3)
                ])

        ret = []

        ret.append(assign_parameter("give_bare_mass", "no"))
        ret.append(assign_parameter("par_b", self.par_b))
        ret.append(assign_parameter("target_m", self.momenta_plus, "_plus", True))
        ret.append(
            assign_parameter("par_m", self.momenta_minus, "_minus", True)
        )
        ret.append(assign_parameter("par_S", self.S_plus, "_plus", True))
        ret.append(assign_parameter("par_S", self.S_minus, "_minus", True))
        ret.append(
            assign_parameter("center_offset", self.center_offset, "", True)
        )

        return "\n".join(ret)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['jhuki', 'jhuki.externals', 'jhuki.externals.nrpypn']

package_data = \
{'': ['*']}

install_requires = \
['ConfigArgParse>=1.2.3,<2.0.0', 'sympy>=1.7.1,<2.0.0']

setup_kwargs = {
    'name': 'jhuki',
    'version': '0.1.0.dev0',
    'description': 'Utilities to prepare Einstein Toolkit simulations',
    'long_description': '<p align="center">\n<img src="https://github.com/Sbozzolo/jhuki/raw/master/logo.png" height="250">\n</p>\n\n[![GPLv3\nlicense](https://img.shields.io/badge/License-GPLv3-blue.svg)](http://perso.crans.org/besson/LICENSE.html)\n![Tests](https://github.com/Sbozzolo/Jhuki/workflows/Tests/badge.svg)\n\n\nWriting parameter files for Cactus simulations is not easy. To achieve a\nsuccessful evolution, one has to tune several parameters, which typically depend\non the grid configuration or on other settings. Writing these par files by hand\nis tedious and error-prone. `Jhuki` is a Python library to prepare\nsimulations for the `Einstein Toolkit` (or Cactus-based codes).\n\n> :warning: This package is currently under development. It may be full of bugs,\n>           and its interfaces might change without notice.\n\n## Features\n\n* Generate parameter files from a template using configuration files or\n  command-line arguments\n* Take care of the grid configuration given the desired resolution at the finest\n  level and other details\n\n## Examples\n\n### Working with grid structures\n\nProblem: you want to generate the parfile code for a grid structure with two\nrefinement centers, each with 8 levels and resolution at the finest level 0.001\nand CFL factor of 0.4 (in the finest level). In this, you want to ensure that\nthe maximum timestep on the grid never exceeds 1 to avoid numerical instability.\n\n``` python\n#!/usr/bin/env python3\n\nfrom jhuki import grid as pg\n\nrefinement_radii = tuple(2**level for level in range(7))\n\ncenter1 = pg.RefinementCenter(refinement_radii,\n                              dx_fine=0.001,\n                              cfl_fine=0.5,\n                              center_num=1,\n                              position=(10,0,0))\n\n# Same but with different center_num and position\ncenter2 = pg.RefinementCenter(refinement_radii,\n                              dx_fine=0.001,\n                              cfl_fine=0.5,\n                              center_num=2,\n                              position=(-10,0,0))\n\ngrid_not_synced = pg.Grid((center1, center2), outer_boundary=1000)\ngrid_synced = pg.set_dt_max_grid(grid_not_synced, dt_max=1)\n\nprint(grid_synced.parfile_code)\n```\nThis will output\n\n``` sh\nCartGrid3D::type = "coordbase"\nCarpet::domain_from_coordbase = "yes"\nCoordBase::domainsize = "minmax"\nCoordBase::xmin = 1000\nCoordBase::ymin = 1000\nCoordBase::zmin = 1000\nCoordBase::xmax = 1000\nCoordBase::ymax = 1000\nCoordBase::zmax = 1000\nCoordBase::dx = 0.64\nCoordBase::dy = 0.64\nCoordBase::dz = 0.64\nCarpet::max_refinement_levels = 8\nCarpet::time_refinement_factors = "[1,1,2,4,8,16,32,64]"\nCarpetRegrid2::num_levels_1 = 8\nCarpetRegrid2::position_x_1 = 10\nCarpetRegrid2::position_y_1 = 0\nCarpetRegrid2::position_z_1 = 0\nCarpetRegrid2::radius_1[1] = 64\nCarpetRegrid2::radius_1[2] = 32\nCarpetRegrid2::radius_1[3] = 16\nCarpetRegrid2::radius_1[4] = 8\nCarpetRegrid2::radius_1[5] = 4\nCarpetRegrid2::radius_1[6] = 2\nCarpetRegrid2::radius_1[7] = 1\nCarpetRegrid2::num_levels_2 = 8\nCarpetRegrid2::position_x_2 = -10\nCarpetRegrid2::position_y_2 = 0\nCarpetRegrid2::position_z_2 = 0\nCarpetRegrid2::radius_2[1] = 64\nCarpetRegrid2::radius_2[2] = 32\nCarpetRegrid2::radius_2[3] = 16\nCarpetRegrid2::radius_2[4] = 8\nCarpetRegrid2::radius_2[5] = 4\nCarpetRegrid2::radius_2[6] = 2\nCarpetRegrid2::radius_2[7] = 1\n```\n\nYou can also add a small shift to the grid so that the origin is not on (0,0,0)\npassing the `tiny_shift` argument to `Grid`.\n\n## Installation\n\nThe best way to install `Jhuki` is by cloning this repo and using\n[poetry](https://python-poetry.org/). If you have poetry install, just run\n`poetry install` in the folder where you cloned the repo to install `Jhuki`.\n\n## Tests\n\n`Jhuki` comes with a suite of unit tests. To run the tests,\n```sh\npoetry run pytest --cov=./ --cov-report=term\n```\nTests are automatically run after each commit by GitHub Actions. This will also\ntell you what is the test coverage.\n\n# Code style\n\n- We lint the code with `black -l 79`.\n\n# What does _jhuki_ mean?\n\nThe word _jhuki_ belongs to the Tohono O\'odham vocabulary and means *rain*. If\n[kuibit](https://githum.com/Sbozzolo/kuibit) is the tool you use to collect the\nfruit of your `Cactus` simulations, then `jhuki` is what allowed that fruit to\nbe there in the first place.\n\n## Credits\n\nThe logo contains elements designed by [pngtree.com](pngtree.com).\n\nThe computation of the momenta for quasi-circular mergers of binary black holes\nuses\n[NRPyPN](https://einsteintoolkit.org/thornguide/EinsteinInitialData/NRPyPN/documentation.html).\nIf you use this module, please follow the citation guidelines as specified by\nthe documentation in the `NRPyPN` repo.\n',
    'author': 'Gabriele Bozzola',
    'author_email': 'gabrielebozzola@arizona.edu',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/sbozzolo/Jhuki',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.6,<4.0',
}


setup(**setup_kwargs)



#!/usr/bin/env python3

# Copyright (C) 2021 Gabriele Bozzola
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation; either version 3 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, see <https://www.gnu.org/licenses/>.

from pytest import approx

import jhuki.externals.nrpypn as nrpypn

# Check that we can use eval__P_t__and__P_r


def test_compute_quasicircular_momenta():
    assert nrpypn.compute_quasicircular_momenta(
        1, 12, (0.1, 0.2, 0.3), (-0.1, -0.2, -0.3)
    ) == approx((0.0850933618449702, 0.000539545145231776))
